# -*- coding: utf-8 -*-
"""
---------------------------------------------
Copyright (c) 2025 ZhangYundi
Licensed under the MIT License. 
Created on 2025/6/25 17:04
Email: yundi.xxii@outlook.com
Description: 注册获取数据的具体实现方法
---------------------------------------------
"""

from typing import Callable

import polars as pl

import quda

FUNC_SQL_LOCAL: Callable[[str, str], pl.LazyFrame] = lambda date, ds_path: quda.sql(
    f"select * from {ds_path} where date='{date}';")



# -*- coding: utf-8 -*-
"""
---------------------------------------------
Copyright (c) 2025 ZhangYundi
Licensed under the MIT License.
Created on 2025/6/25 15:05
Email: yundi.xxii@outlook.com
Description: 主流程控制
---------------------------------------------
"""
from collections.abc import Sequence

import mlflow
import ylog
from hydra import compose, initialize
from hydra.utils import instantiate
from omegaconf import DictConfig
from sklearn.base import BaseEstimator
from sklearn.model_selection import train_test_split
from sklearn.pipeline import Pipeline

import quda
from .data import DataLoader
from .pipeline import IncrementalPipeline
from .trainer import BaseIncrementalTrainer
from .utils import extract_feature_names, get_timedelta_names


class WorkFlow(BaseEstimator):
    """
    构建 训练工作流: 从数据加载->预处理->增量训练。支持 sklearn式更改配置、从配置文件中加载
    """

    def __init__(self,
                 data: list,
                 trainer: BaseIncrementalTrainer,
                 preprocessor: list | None = None,
                 trainer_params: dict | None = None):
        """
        构造工作流

        Parameters
        ----------
        data: list
            数据加载步骤
        trainer: BaseIncrementalTrainer
            训练器：quda.ml.trainer.BaseIncrementalTrainer 的实现实例
        preprocessor: list
            预处理步骤，默认None: 不做预处理
        trainer_params:
            训练器参数
        """
        # 初始化
        self.data = DataLoader(steps=data)
        self.trainer = trainer
        if isinstance(self.trainer, type):
            self.trainer = self.trainer()
        assert isinstance(self.trainer,
                          BaseIncrementalTrainer), "trainer must be a subclass of `BaseIncrementalTrainer`"
        self.trainer_params = trainer_params
        self.preprocessor = IncrementalPipeline(preprocessor) if preprocessor else None
        self.pipe = Pipeline([
            ("data", self.data.pipe),
            ("trainer", self.trainer)
        ])
        if self.preprocessor:
            self.pipe.steps.insert(1, ("preprocessor", self.preprocessor.pipe))

        self._html_repr = self.pipe._html_repr
        # 配置文件中解析出来的参数: 包括 workflow 配置参数，数据集配置参数，模型配置参数，训练参数，fit 运行配置
        self._cfg: DictConfig | None = None

    @classmethod
    def from_conf(cls, config_path: str, config_name: str) -> 'WorkFlow':
        """
        配置驱动:从配置文件中构建工作流

        Examples
        --------
        >>> import quda.ml
        >>> workflow = quda.ml.from_conf(config_path='./conf', config_name="config")
        >>> workflow.set_params(...) # 更改配置
        >>> workflow.fit(...) # 开始训练

        Returns
        -------
        """

        with initialize(version_base=None, config_path=config_path):
            cfg = compose(config_name=config_name)

        # 构建 data pipeline
        data_steps = []
        data_steps = instantiate(cfg.data)
        data_steps = [(name, step) for d in data_steps for name, step in d.items()]

        # 构建 preprocessor
        preprocessor_steps = None
        if cfg.get("preprocessor"):
            preprocessor_steps = []
            preprocessor_steps = instantiate(cfg.preprocessor)
            preprocessor_steps = [(name, step) for d in preprocessor_steps for name, step in d.items()]

        # 构建 trainer
        trainer = instantiate(cfg.trainer)

        # trainer_params
        trainer_params = instantiate(cfg.trainer_params)

        wf = cls(data=data_steps,
                 trainer=trainer,
                 preprocessor=preprocessor_steps,
                 trainer_params=trainer_params)
        wf._cfg = cfg
        return wf

    def fit(self,
            iterables: Sequence,
            batch_size: int = -1,
            features: list[str] | None = None,
            valid_size: float = 0.2,
            shuffle: bool = False,
            random_state: int = 42, ):
        """
        通用增量训练流程，支持分批加载数据,预处理，模型训练与保存

        Parameters
        ----------
        iterables: Sequence
            可迭代对象
        batch_size: int
            每次训练的批次，默认 -1: 全量训练
        features: list[str]|None
            特征名, 默认None。会排除date,time,asset,price,以timedelta命名的收益列
        valid_size: float
            验证集比例，默认0.2
        shuffle: bool
            是否打乱 iterables, 默认 False
        random_state: int
            随机数种子，默认 42
        Returns
        -------

        """
        batch_size = batch_size if batch_size > 0 else len(iterables)
        iterables = sorted(iterables)

        with mlflow.start_run() as top_run:
            top_run_id = top_run.info.run_id
            ylog.info(f"Started mlflow.run: {top_run_id}")
            params = {
                "data.beg": iterables[0],
                "data.end": iterables[-1],
                "batch_size": batch_size,
                "shuffle": shuffle,
                "random_state": random_state,
            }
            for i, train_batch in enumerate(
                    [iterables[i:i + batch_size] for i in range(0, len(iterables), batch_size)]):
                beg, end = train_batch[0], train_batch[-1]
                evals_result = {}
                with mlflow.start_run(nested=True, log_system_metrics=True) as partial_run:
                    partial_run_id = partial_run.info.run_id
                    ylog.info(f"Started mlflow.run: {partial_run_id}")
                    ylog.info(f"Preparing on batch_{i + 1} ({i + 1}/{len(iterables)}) - {beg} >>> {end}")
                    params["data.beg"] = beg
                    params["data.end"] = end
                    # 加载训练数据
                    train_list_, valid_list_ = train_test_split(train_batch,
                                                                test_size=valid_size,
                                                                shuffle=shuffle,
                                                                random_state=random_state)
                    train_dl = self.data.fetch(train_list_, batch_size=batch_size, show_progress=False)
                    train_data = next(train_dl)
                    if self.preprocessor:
                        train_data = self.preprocessor.partial_fit_transform(train_data)
                    if valid_list_:
                        valid_dl = self.data.fetch(valid_list_, batch_size=batch_size, show_progress=False)
                        valid_data = next(valid_dl)
                        if self.preprocessor:
                            valid_data = self.preprocessor.fit_transform(valid_data)
                    else:
                        valid_data = None

                    if not features:
                        features = extract_feature_names(train_data.columns)
                    else:
                        features = features
                    if self.trainer.input_example is None:
                        self.trainer.input_example = train_data[features].head().to_pandas()

                    target = get_timedelta_names(train_data.columns)
                    if len(target) > 1:
                        ylog.warning(f"Multiple targets found: {target}, use {target[0]}")
                    else:
                        raise ValueError("Miss target columns in data")

                    target = target[0]

                    ylog.info("Building dataset.train...")
                    train_data = self.trainer.create_dataset(data=train_data,
                                                             features=features,
                                                             target=target, )

                    if valid_data:
                        ylog.info("Building dataset.valid...")
                        valid_data = self.trainer.create_dataset(data=valid_data,
                                                                 features=features,
                                                                 target=target, )

                    ylog.info(f"Training on batch_{i + 1} ({i + 1}/({len(iterables)}) - {beg} >>> {end}")

                    # 训练模型
                    model = self.trainer.fit(train_data=train_data,
                                             valid_data=valid_data,
                                             init_model=self.trainer.model,
                                             evals_result=evals_result,
                                             **self.trainer.params)

                    mlflow.log_metrics(evals_result, run_id=partial_run_id)

                    # 保存模型
                    model_dir = quda.tb_path("models")
                    model_dir.mkdir(parents=True, exist_ok=True)
                    artifact_path = model_dir / f"checkpoint_{i + 1}"
                    model_info = self.trainer.save_model(model, artifact_path, )
                    self.trainer.model = mlflow.pyfunc.load_model(model_info.model_uri)

                    model = None

    @classmethod
    def run(cls, config_path: str, config_name: str) -> 'WorkFlow':
        """
        从配置文件中创建 WorkFlow 对象，并且运行 fit
        Returns
        -------

        """
        wf = cls.from_conf(config_path=config_path, config_name=config_name)
        fit_config = wf._cfg.get("fit")
        if not fit_config:
            ylog.warning("Missing WorkFlow.fit params.")
            return
        return cls.fit(**fit_config)


from_conf = WorkFlow.from_conf
run = WorkFlow.run

def init_config():
    """初始化配置文件."""
    from pathlib import Path
    import shutil
    from importlib.resources import files

    source_conf_dir = files("quda.ml").joinpath("conf")
    # 当前工作目录下的 conf/
    target_conf_dir = Path.cwd() / "conf"
    if target_conf_dir.exists():
        ylog.warning(f"[QUDA-ML] - {target_conf_dir} already exists.")
        return
    try:
        shutil.copytree(source_conf_dir, target_conf_dir)
        ylog.info(f"[QUDA-ML] - {target_conf_dir} created.")
    except Exception as e:
        ylog.error(f"[QUDA-ML] - Failed to create {target_conf_dir}\n{e}")


# -*- coding: utf-8 -*-
"""
---------------------------------------------
Copyright (c) 2025 ZhangYundi
Licensed under the MIT License.
Created on 2025/6/25 15:05
Email: yundi.xxii@outlook.com
Description:
---------------------------------------------
"""

from collections.abc import Sequence

def get_timedelta_names(columns: Sequence[str], ) -> list[str]:
    """
    Return a list of column names that can be successfully parsed into <pandas>.Timedelta object.

    Parameters
    ----------
    columns: Sequence[str]
        A Sequence of strings representing column names to be evaluated for Timedelta

    Returns
    -------
    list[str]

    Examples
    --------
    >>> get_timedelta_names(['1s', '2min', 'invalid', '3H'])
    ['1s', '2min', '3H']

    """
    from pandas import Timedelta
    def _is_valid_timedelta(s: str) -> bool:
        try:
            Timedelta(s)
            return True
        except ValueError:
            return False

    return [col for col in columns if _is_valid_timedelta(col)]


def extract_feature_names(columns: Sequence[str], ) -> list[str]:
    """
    Extract a list of valid feature column names by excluding index-like(date/time/asset)、price-like(price)、time-like columns.
    Parameters
    ----------
    columns: Sequence[str]
        A sequence of strings representing all available column names.

    Returns
    -------
    list[str]

    Examples
    --------
    >>> extract_feature_names(['date', 'time', 'asset', 'open', 'high', '1s', 'volume', 'price'])
    ['open', 'high', 'volume']

    """
    return_cols = get_timedelta_names(columns)
    exclude = {"date", "time", "asset", 'price', *return_cols}
    return [col for col in columns if col not in exclude]


# -*- coding: utf-8 -*-
"""
---------------------------------------------
Copyright (c) 2025 ZhangYundi
Licensed under the MIT License.
Created on 2025/6/25 15:05
Email: yundi.xxii@outlook.com
Description: 增量流水线
---------------------------------------------
"""

from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.pipeline import Pipeline


class IncrementalPipeline(BaseEstimator, TransformerMixin):
    """
    支持 partial_fit 的 Pipeline。
    只对支持 partial_fit 的 transformer 执行该操作，其余保存不变。
    """

    def __init__(self, steps: list):
        pipe_steps = list()
        if steps:
            for name, step in steps:
                if isinstance(step, type):
                    pipe_steps.append((name, step()))
                else:
                    pipe_steps.append((name, step))
        self.pipe = Pipeline(pipe_steps)

    def fit(self, X, y=None):
        """完整fit整个pipeline"""
        for _, step in self.pipe.steps:
            X = step.fit_transform(X, y)
        return self

    def transform(self, X):
        for _, step in self.pipe.steps:
            X = step.transform(X)
        return X

    def fit_transform(self, X, y=None, **fit_params):
        return self.fit(X, y).transform(X)

    def partial_fit(self, X, y=None):
        """只对支持 partial_fit 的步骤进行增量更新"""
        for name, step in self.pipe.steps:
            if hasattr(step, "partial_fit"):
                step.partial_fit(X, y)
            else:
                step.fit(X, y)
            X = step.transform(X)
        return self

    def partial_fit_transform(self, X, y=None):
        return self.partial_fit(X, y).transform(X)

    def set_params(self, **params):
        """允许通过类似 sklearn 的 params__value设置参数"""
        self.pipe.set_params(**params)
        return self

    def get_params(self, deep=True):
        """返回所有步骤的参数，用于 mlflow logging 或者 gridsearch cv"""
        return self.pipe.get_params(deep=deep)



# -*- coding: utf-8 -*-
"""
---------------------------------------------
Created on 2025/6/23 20:34
@author: ZhangYundi
@email: yundi.xxii@outlook.com
@description: 
---------------------------------------------
"""
import lightgbm as lgb
import mlflow
import ylog

from .base import BaseIncrementalTrainer


class LightGBMTrainer(BaseIncrementalTrainer):

    def __init__(self, experiment_name: str = "LightGBM-Incremental",):
        super().__init__(experiment_name)

    def create_dataset(self, data, features: list[str], target: str):
        return lgb.Dataset(data[features].to_numpy(), label=data[target].to_numpy(), feature_name=features)

    def fit(self,
            train_data: lgb.Dataset,
            valid_data: lgb.Dataset | None = None,
            init_model=None,
            evals_result: dict | None = None,
            **model_params):
        evals_result = dict() if evals_result is None else evals_result

        # 检查 参数
        params = {"early_stopping_rounds": 10,
                  "num_boost_round": 100,
                  "log_eval_period": 10}
        for check in params.keys():
            if check not in model_params:
                ylog.warning(f"Missing `{check}`, set default vlaue: {params.get(check)}")
            else:
                params[check] = model_params.pop(check)

        early_stopping_callback = lgb.early_stopping(params["early_stopping_rounds"])
        verbose_eval_callback = lgb.log_evaluation(params["log_eval_period"])
        callbacks = [early_stopping_callback, verbose_eval_callback, self._mlflow_callback(evals_result)]

        valid_sets = [train_data, ]
        valid_names = ["train"]
        if valid_data:
            valid_sets.append(valid_data)
            valid_names.append("valid")

        model = lgb.train(
            model_params,
            train_set=train_data,
            init_model=init_model,
            valid_sets=valid_data,
            valid_names=valid_names,
            num_boost_round=model_params.pop("num_boost_round"),
            callbacks=callbacks,
        )

        return model

    def _mlflow_callback(self, evals_result):
        """mlflow 回调"""

        def _callback(env):
            for data_name, eval_name, result, _ in env.evaluation_result_list:
                name = f"{data_name}.{eval_name}".replace("@", "_")
                metrics = {name: result}
                mlflow.log_metrics(metrics, step=env.iteration)
                evals_result.update({name: result})

        return _callback



# -*- coding: utf-8 -*-
"""
---------------------------------------------
Created on 2025/6/23 15:09
@author: ZhangYundi
@email: yundi.xxii@outlook.com
@description: 
---------------------------------------------
"""

from .base import BaseIncrementalTrainer
from .lgbm import LightGBMTrainer


# -*- coding: utf-8 -*-
"""
---------------------------------------------
Copyright (c) 2025 ZhangYundi
Licensed under the MIT License. 
Created on 2025/6/25 10:20
Email: yundi.xxii@outlook.com
Description: 
---------------------------------------------
"""

from .data import DataLoader
from .workflow import WorkFlow, from_conf, run, init_config

__version__ = "0.1.3"

__all__ = [
    "WorkFlow",
    "DataLoader",
    "from_conf",
    "run",
    "init_config"
]




# -*- coding: utf-8 -*-
"""
---------------------------------------------
Created on 2025/6/23 09:49
@author: ZhangYundi
@email: yundi.xxii@outlook.com
@description: 基础转换器仓库
---------------------------------------------
"""

import inspect
import math
from collections.abc import Sequence
from typing import Literal

import numpy as np
import polars as pl
import ygo
from pandas import Timedelta
from sklearn.base import BaseEstimator, TransformerMixin

SPECIAL_VALUES = {
    "null": None,
    "none": None,
    "nan": None,
    "inf": np.inf,
    "-inf": -np.inf,
}


class Function(BaseEstimator, TransformerMixin):
    """
    将任意函数封装为支持 set_params 的 BaseEstimator

    Parameters
    ----------
    fn: callable
        方法对象
    kwargs: dict
        fn的参数

    Examples
    --------
    >>> import quda
    >>> func = Function(lambda date, ds_path: quda.sql(f"select * from {ds_path} where date='{date}'"))
    >>> func.set_params(ds_path="mc/stock_kline_day")
    """

    def __init__(self, fn: callable, **kwargs):
        self.fn = ygo.delay(fn)(**kwargs)
        self.params = {k: inspect.Parameter.empty for k in ygo.fn_signature_params(self.fn)}
        self.params.update(self.fn.stored_kwargs)

    def set_params(self, **params):
        self.fn = ygo.delay(self.fn)(**params)
        self.params.update(self.fn.stored_kwargs)
        return self

    def get_params(self, deep=True):
        return self.params

    def fit(self, X, y=None):
        return self

    def transform(self, X):
        if self.fn is None:
            raise ValueError("Function not provided.")
        params = {k: v for k, v in self.fn.stored_kwargs.items() if v is not inspect.Parameter.empty}
        return self.fn(X, **params)


class FilterNotNull(BaseEstimator, TransformerMixin):
    """
    Drop all rows that contain one or more null values

    Parameters
    ----------
    subset: Sequence[str] | str | None
        剔除空值的字段

    """

    def __init__(self, subset: Sequence[str] | str | None = None):
        self.subset = subset

    def fit(self, X=None, y=None):
        return self

    def transform(self, X: pl.LazyFrame):
        if self.subset:
            return X.drop_nulls(self.subset)
        return X


class Cast(BaseEstimator, TransformerMixin):
    """
    类型转换

    Parameters
    ----------
    old: str | polars.DataType | polars.Expr
        原本类型或者需要转换的列
    new: polars.DataType
        转换的目标类型

    Examples
    --------
    >>> Cast("asset", pl.UInt16)

    """

    def __init__(self, old: pl.Expr | str | pl.DataType = None, new=None):
        self.old = old
        self.new = new
        self.old_expr: pl.Expr = None

    def fit(self, X, y=None):
        if isinstance(self.old, str):
            if self.old in SPECIAL_VALUES:
                self.old = SPECIAL_VALUES[self.old]
        self.old_expr = self.old if isinstance(self.old, pl.Expr) else pl.col(self.old_expr)
        return self

    def transform(self, X):
        return X.with_columns(self.old_expr.cast(self.new))


class Expression(BaseEstimator, TransformerMixin):
    """
    添加新列

    Parameters
    ----------
    exprs: list[polars.Expr]
        需要添加的列(使用`polars.Expr`表示)

    """

    def __init__(self, exprs: list[pl.Expr] | None = None):
        self.exprs = exprs

    def fit(self, X, y=None):
        return self

    def transform(self, X):
        for expr in self.exprs:
            X = X.with_columns(expr)
        return X


class Imputer(BaseEstimator, TransformerMixin):
    """
    插值器

    Parameters
    ----------
    strategy
        插值策略
    over_spec
        用于滚动插值
    columns
        需要处理的列
    """

    def __init__(self,
                 strategy: Literal[
                               "forward", "backward", "mean", "zero", "max", "min", "one"] | None = "forward",
                 over_spec: dict | None = None,
                 columns: Sequence[str] | None = None):
        self.strategy = strategy
        self.over_spec = over_spec
        self.columns = [columns, ] if isinstance(columns, str) else columns
        self._exprs: list[pl.Expr] = list()

    def fit(self, X=None, y=None):
        return self

    def transform(self, X: pl.LazyFrame):
        cols = X.collect_schema().names() if not self.columns else self.columns
        if not self.over_spec:
            self._exprs = [pl.col(col).fill_null(self.strategy) for col in cols]
        else:
            self._exprs = [pl.col(col).fill_null(self.strategy).over(**self.over_spec) for col in self.columns]
        return X.with_columns(self._exprs)


class Replace(BaseEstimator, TransformerMixin):
    """
    替换器

    Parameters
    ----------
    columns: list
        需要替换的目标列
    old
        旧值, 可以是list, 如[np.inf, -np.inf]
    new
        新值
    """

    def __init__(self, columns: list | None = None, old=None, new=None):

        self.columns = columns
        self.old = old
        self.new = new

    def fit(self, X, y=None):
        if isinstance(self.old, str):
            if self.old in SPECIAL_VALUES:
                self.old = SPECIAL_VALUES[self.old]
        elif isinstance(self.old, list):
            self.old = [SPECIAL_VALUES[old] if old in SPECIAL_VALUES else old for old in self.old]
        if self.columns is None:
            self.columns = list()
        return self

    def transform(self, X):
        return X.with_columns(pl.col(c).replace(self.old, self.new) for c in self.columns)


class Target(BaseEstimator, TransformerMixin):
    """
    添加目标收益列

    Parameters
    ----------
    price_tag
        价格列，支持sql语法，如 `if(ask_p1 > 0, bid_p1 > 0, (ask_p1+bid_p1)/2, null) as price`
    frequency
        数据集频率, 默认`3s`
    target
        目标收益，默认`5min`
    partition_by
        分区
    order_by
        排序
    """

    def __init__(self,
                 price_tag: str = "price",
                 frequency: str = "3s",
                 target: str = "5min",
                 partition_by: Sequence[str] | str = "asset",
                 order_by: Sequence[str] | str | None = "time", ):
        self.price_tag = price_tag
        self.frequency = frequency
        self.target = target
        self.partition_by = partition_by
        self.order_by = order_by

    def fit(self, X=None, y=None):
        return self

    def transform(self, X: pl.LazyFrame):
        over_spec = {"partition_by": self.partition_by}
        if self.order_by:
            over_spec["order_by"] = self.order_by
        freq_secs = int(Timedelta(self.frequency).seconds)
        expr_target = (
            (
                    pl.sql_expr(self.price_tag)
                    .shift(-math.ceil(Timedelta(self.target).seconds / freq_secs))
                    .over(**over_spec) / pl.sql_expr(self.price_tag) - 1
            ).alias(self.target)
        )
        expr_inf = pl.col(self.target).replace([float("inf"), float("-inf")], None)
        exprs = [expr_target, expr_inf]
        for expr in exprs:
            X = X.with_columns(expr)
        return X.fill_nan(None)



# -*- coding: utf-8 -*-
"""
---------------------------------------------
Created on 2025/6/23 09:54
@author: ZhangYundi
@email: yundi.xxii@outlook.com
@description: 支持 partial_fit(增量fit)
---------------------------------------------
"""

from sklearn.base import BaseEstimator, TransformerMixin
import polars as pl

class StandardScaler(BaseEstimator, TransformerMixin):

    """
    Parameters
    ----------
    ddof: int
        自由度(0为总体标准差，1为样本标准差)
    """

    def __init__(self, subset: list[str], ddof=1):
        self.subset = subset
        self._count = {col: 0 for col in subset}
        self._mean = {col: 0 for col in subset}
        self._M2 = {col: 0 for col in subset}
        self.count = None
        self.mean = None
        self.M2 = None
        self.ddof = ddof
        self.reset()

    def reset(self):
        """重置"""
        self._count = {col: 0 for col in self.subset}
        self._mean = {col: 0 for col in self.subset}
        self._M2 = {col: 0 for col in self.subset}
        self.count = pl.DataFrame(self._count)
        self.mean = pl.DataFrame(self._mean)
        self.M2 = pl.DataFrame(self._M2)

    def fit(self, X: pl.DataFrame, y=None):
        self.reset()
        self.partial_fit(X, y)
        return self

    def partial_fit(self, X: pl.DataFrame, y=None):
        data = X[self.subset]
        mu_new = data.mean()
        n_new = data.count() - data.null_count()
        m2_new = data.select(((pl.col(c) - mu_new[c])**2).sum() for c in self.subset)
        # 新旧均值差异
        delta = mu_new - self.mean
        # 全局均值
        new_count = self.count + n_new
        new_mean = (self.mean * self.count + mu_new * n_new) / new_count
        # 全局M2
        # 误差项
        correction_term = delta.select(pl.col(c) ** 2 * (self.count[c] * n_new[c]) / new_count[c] for c in self.subset)

        new_M2 = self.M2 + m2_new + correction_term
        # 更新状态
        self.count = new_count
        self.mean = new_mean
        self.M2 = new_M2
        return self

    def get_std(self, col: str):
        return (self.M2[col] / (self.count[col] - self.ddof)) ** 0.5

    def transform(self, X: pl.DataFrame):
        if not self.subset:
            return X
        return (X
                .with_columns((pl.col(c) - self.mean[c])/self.get_std(c) for c in self.subset)
                .with_columns(pl.col(c).replace([float("inf"), float("-inf")], None) for c in self.subset))

    def partial_fit_transform(self, X: pl.DataFrame, y=None):
        if not self.subset:
            return X
        return self.partial_fit(X).transform(X)

